/////////////////////////
// convert_recording.c //
/////////////////////////

/******************************************************************************
 * Converts an uncompressed binary recording to an ASCII text file that can
 * be read into MATLAB or similar.
******************************************************************************/

/******************************************************************************
 * Copyright  2011, 2012 Mike Roddewig (mroddewi@mines.edu).
 * All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License v3 as published 
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
******************************************************************************/

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <argp.h>
#include <string.h>

#include "version.h"

/////////////
// Defines //
/////////////

const char * program_name = "convert_recording";

////////////////
// argp stuff //
////////////////

/* This structure is used by main to communicate with parse_opt. */
struct arguments {
	char * output_filename;
	char * input_filename;
};


const char *argp_program_version = CONVERT_RECORDING_VERSION;

const char *argp_program_bug_address = "Mike Roddewig <mroddewi@mines.edu>";

/*
   OPTIONS.  Field 1 in ARGP.
   Order of fields: {NAME, KEY, ARG, FLAGS, DOC}.
*/
static struct argp_option options[] = {
	{"output", 'o', "FILENAME", 0,
		"Output to file FILENAME."},
	{0}
};


/*
   PARSER. Field 2 in ARGP.
   Order of parameters: KEY, ARG, STATE.
*/

static error_t parse_opt (int key, char *arg, struct argp_state *state) {
	struct arguments *arguments = (struct arguments *) state->input;

	switch (key) {
		case ARGP_KEY_ARG:
			if (state->arg_num >= 1) {
				argp_usage(state);
			}
			arguments->input_filename = arg;
			break;
		case 'o':
			arguments->output_filename = arg;
			break;
		case ARGP_KEY_END:
			break;
		default:
		return ARGP_ERR_UNKNOWN;
	}
	
	return 0;
}

/*
   ARGS_DOC. Field 3 in ARGP.
   A description of the non-option command-line arguments
     that we accept.
*/
static char args_doc[] = "INPUT";

/*
  DOC.  Field 4 in ARGP.
  Program documentation.
*/
static char doc[] = "Convert the INPUT recording file to a text file, delimited with commas.";

/*
   The ARGP structure itself.
*/
static struct argp argp = {options, parse_opt, args_doc, doc};

//////////
// Main //
//////////

int main(int argc, char * argv[]) {
	struct arguments arguments;
	int8_t ret;
	FILE * input_file;
	FILE * output_file;
	int8_t done;
	double sample;
	
	char * output_filename;
	
	// Handle any command-line arguments.
	
	arguments.input_filename = NULL;
	arguments.output_filename = NULL;
	
	argp_parse(&argp, argc, argv, 0, 0, &arguments);
	
	if (arguments.input_filename == NULL) {
		fprintf(stderr, "You must specify the input filename.\n");
		exit(EXIT_FAILURE);
	}
	
	input_file = fopen(arguments.input_filename, "r");
	
	if (input_file == NULL) {
		fprintf(stderr, "Could not open the input file.\n");
		exit(EXIT_FAILURE);
	}
	
	if (arguments.output_filename == NULL) {
		output_filename = (char *) malloc(strlen(arguments.input_filename) + 5);
		sprintf(output_filename, "%s.txt", arguments.input_filename);
	}
	else {
		output_filename = arguments.output_filename;
	}
	
	output_file = fopen(output_filename, "w");
	
	if (output_file == NULL) {
		fprintf(stderr, "Could not open the output file.\n");
		exit(EXIT_FAILURE);
	}
	
	done = 0;
	
	while (done == 0) {
		ret = fread(&sample, sizeof(double), 1, input_file);
		
		if (ret != 1) {
			if (feof(input_file) != 0) {
				done = 1;
			}
			else {
				fprintf(stderr, "Error reading from the input file.\n");
				done = 1;
			}
		}
		else {
			ret = fprintf(output_file, "%f\n", sample);
			
			if (ret == 0) {
				fprintf(stderr, "Error writing to the output file.\n");
				done = 1;
			}
		}
	}
	
	fclose(input_file);
	fclose(output_file);
	
	return EXIT_SUCCESS;
}