/////////////////////
// telescope_audio //
/////////////////////

/******************************************************************************
 * Copyright  2011, 2012 Mike Roddewig (mroddewi@mines.edu).
 * All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License v3 as published 
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
******************************************************************************/

#include <syslog.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include <stdio.h>
#include <argp.h>
#include <unistd.h>
#include <complex>

#include <sndfile.h>

#include "uhd/types/device_addr.hpp"
#include "uhd/types/ref_vector.hpp"
#include "uhd/usrp/multi_usrp.hpp"
#include "uhd/device.hpp"
#include <uhd/utils/thread_priority.hpp>
#include <uhd/utils/safe_main.hpp>
#include <uhd/utils/msg.hpp>

#include <boost/thread.hpp>

#include "telescope.h"
#include "version.h"

using namespace uhd;
using namespace uhd::usrp;

/////////////
// Defines //
/////////////

#define USRP_TIMEOUT		0.1	// sec
#define USRP_SAMPLE_RATE	200e3	// S/s
#define AUDIO_SAMPLE_RATE	44.1e3	// S/s
#define USRP_RX_FREQ		20.1e6	// Hz
#define USRP_GAIN		0	// dB
#define STARTUP_TIME		1.5	// sec
#define BUFFER_SIZE		1024	// samples
#define INTEGRATION_PERIOD	1	// seconds
#define SCALING_GAIN		1e3

////////////////
// argp stuff //
////////////////

const char * program_name = "telescope_audio";

/* This structure is used by main to communicate with parse_opt. */
struct arguments {
	char * address;
	char * audiofile;
	uint8_t verbose;
	double sample_rate;
	double rx_freq;
	double gain;
	double period;
};


const char *argp_program_version = TELESCOPE_AUDIO_VERSION;

const char *argp_program_bug_address = "Mike Roddewig <mroddewi@mines.edu>";

/*
   OPTIONS.  Field 1 in ARGP.
   Order of fields: {NAME, KEY, ARG, FLAGS, DOC}.
*/
static struct argp_option options[] = {
	{"audio", 'a', "AUDIOFILE", 0,
		"The path of the output audio file."},
	{"samplerate", 's', "SAMPLERATE", OPTION_ARG_OPTIONAL, 
		"Audio sample rate after decimation, in SPS. Default is 44.1 kHz."},
	{"rxfreq", 'r', "RXFREQ", OPTION_ARG_OPTIONAL,
		"Receiver frequency to tune to, in Hz. Default is 20.1 MHz."},
	{"gain", 'g', "GAIN", OPTION_ARG_OPTIONAL,
		"Receiver gain in dB. Default is 0 dB."},
	{"verbose", 'v', 0, OPTION_ARG_OPTIONAL,
		"Turn on extra messages."},
	{0}
};


/*
   PARSER. Field 2 in ARGP.
   Order of parameters: KEY, ARG, STATE.
*/

static error_t parse_opt (int key, char *arg, struct argp_state *state) {
	struct arguments *arguments = (struct arguments *) state->input;
	int ret;

	switch (key) 
	{
		case ARGP_KEY_ARG:
                        if (state->arg_num >= 1) {
                                argp_usage(state);
                        }
                        arguments->address = arg;
                        break;
		case 'a':
			arguments->audiofile = arg;
			break;
		case 'v':
			arguments->verbose = 1;
			break;
		case 's':
			ret = sscanf(arg, "%lf", &arguments->sample_rate);
			if (ret != 1)
			{
				return ARGP_ERR_UNKNOWN;
			}
			break;
		case 'g':
			ret = sscanf(arg, "%lf", &arguments->gain);
			if (ret != 1)
			{
				return ARGP_ERR_UNKNOWN;
			}
			break;
		case 'r':
			ret = sscanf(arg, "%lf", &arguments->rx_freq);
			if (ret != 1)
			{
				return ARGP_ERR_UNKNOWN;
			}
			break;
		case 'p':
			ret = sscanf(arg, "%lf", &arguments->period);
			if (ret != 1)
			{
				return ARGP_ERR_UNKNOWN;
			}
			break;
		case ARGP_KEY_END:
			break;
		default:
			return ARGP_ERR_UNKNOWN;
	}
	
	return 0;
}

/*
   ARGS_DOC. Field 3 in ARGP.
   A description of the non-option command-line arguments
     that we accept.
*/
static char args_doc[] = "[ADDRESS]";

/*
  DOC.  Field 4 in ARGP.
  Program documentation.
*/
static char doc[] = "Capture audio from the USRP at ADDRESS.";

/*
   The ARGP structure itself.
*/
static struct argp argp = {options, parse_opt, args_doc, doc};

///////////////////////
// Private Functions //
///////////////////////

void initialize_usrp(char *, double, double, double);
void capture_data(SNDFILE *, uint32_t);
void my_uhd_handler(uhd::msg::type_t, const std::string &);
double filter(double, uint32_t);

//////////////////////
// Global Variables //
//////////////////////

multi_usrp::sptr dev;
double sample_time;

const uint32_t BL = 21;
const double B[21] = {
    0.01851748725472,  0.02519163567603,  0.03228145412982,  0.03952251731745,
    0.04662426480865,  0.05328604897471,  0.05921437096986,  0.06414015454996,
    0.06783488398029,  0.07012449549516,  0.07090005779818,  0.07012449549516,
    0.06783488398029,  0.06414015454996,  0.05921437096986,  0.05328604897471,
    0.04662426480865,  0.03952251731745,  0.03228145412982,  0.02519163567603,
    0.01851748725472
};



//////////
// Main //
//////////

int UHD_SAFE_MAIN(int argc, char * argv[]) 
{	
	struct arguments arguments;
	
	SNDFILE * output_stream;
	SF_INFO sndfile_info;
	
	uint32_t decimation;
	
	uhd::msg::register_handler(&my_uhd_handler); // Redirect UHD messages to syslog
	
	set_thread_priority_safe();
	
	// Handle any command-line arguments.
	
	arguments.address = NULL;
	arguments.audiofile = NULL;
	arguments.rx_freq = 0;
	arguments.sample_rate = 0;
	arguments.gain = 0;
	arguments.verbose = 0;
	arguments.period = 0;
	
	argp_parse(&argp, argc, argv, 0, 0, &arguments);
	
	if (arguments.rx_freq == 0)
	{
		arguments.rx_freq = USRP_RX_FREQ;
	}
	
	if (arguments.sample_rate == 0)
	{
		arguments.sample_rate = AUDIO_SAMPLE_RATE;
	}
	
	if (arguments.gain == 0)
	{
		arguments.gain = USRP_GAIN;
	}
		

	openlog(program_name, LOG_PERROR, LOG_USER);

	syslog(LOG_INFO, "starting up.");
	
	if (arguments.verbose == 0)
	{
		setlogmask(~(LOG_MASK(LOG_DEBUG)));
	}
	
	if (arguments.address == NULL) {
		syslog(LOG_ERR, "you must specify the address of the USRP. Quitting.");
		exit(EXIT_FAILURE);
	}

	
	// Compute the decimation factor and the actual sample rate.
	
	decimation = (uint32_t) ceil(USRP_SAMPLE_RATE / arguments.sample_rate);
	arguments.sample_rate *= ((double) decimation);
	
	syslog(LOG_INFO, "USRP sample rate is %lf with a decimation rate of %d.", arguments.sample_rate, decimation);

	
	initialize_usrp(arguments.address, arguments.sample_rate, 
		arguments.rx_freq, arguments.gain);
	
	// Open the sound file for writing
	
	sndfile_info.format = SF_FORMAT_WAV | SF_FORMAT_FLOAT | SF_ENDIAN_FILE;
	sndfile_info.channels = 2;
	sndfile_info.samplerate = (int) (dev->get_rx_rate(0) / decimation);
	
	if (sf_format_check(&sndfile_info) == 0)
	{
		syslog(LOG_ERR, "Invalid audio file parameters! Sample rate was %d.", ((int) (arguments.sample_rate / decimation)));
		exit(EXIT_FAILURE);
	}
	
	output_stream = sf_open(arguments.audiofile, SFM_WRITE, &sndfile_info);
	
	if (output_stream == 0x00)
	{
		syslog(LOG_ERR, "Failed to open the audio file. %s", sf_strerror(output_stream));
		exit(EXIT_FAILURE);
	}
	
	
	capture_data(output_stream, decimation);
	
	exit(EXIT_SUCCESS);
}

void initialize_usrp(char * address, double sample_rate, double rx_freq, double gain)
{
	device_addr_t dev_addr;
	tune_request_t tune_request;
	tune_result_t tune_result;
	
	dev_addr["addr0"] = address;
	
	dev = multi_usrp::make(dev_addr);
	
	dev->set_rx_subdev_spec(subdev_spec_t("A:A A:B"));
	dev->set_rx_rate(sample_rate, multi_usrp::ALL_CHANS);
	
	syslog(LOG_DEBUG, "Channel 0 actual receive rate: %f S/s", dev->get_rx_rate(0));
	syslog(LOG_DEBUG, "Channel 1 actual receive rate: %f S/s", dev->get_rx_rate(1));
	
	dev->set_rx_gain(gain, multi_usrp::ALL_GAINS);
	
	syslog(LOG_DEBUG, "Channel 0 actual gain: %f dB", dev->get_rx_gain(0));
	syslog(LOG_DEBUG, "Channel 1 actual gain: %f dB", dev->get_rx_gain(1));
	
	dev->set_rx_dc_offset(true, multi_usrp::ALL_CHANS);
	//dev->set_rx_dc_offset(std::complex<double> (0, 0), multi_usrp::ALL_CHANS);
	
	dev->set_time_now(uhd::time_spec_t(0.0));
        boost::this_thread::sleep(boost::posix_time::milliseconds(50));	// wait for time sync
	
	tune_request = tune_request_t(rx_freq);
	tune_result = dev->set_rx_freq(tune_request, 0);
	
	syslog(LOG_DEBUG, "Actual DSP0 frequency: %f Hz", tune_result.actual_dsp_freq);
	
	tune_result = dev->set_rx_freq(tune_request, 1);
	
	syslog(LOG_DEBUG, "Actual DSP1 frequency: %f Hz", tune_result.actual_dsp_freq);
	
	syslog(LOG_DEBUG, "USRP configuration: %s", dev->get_pp_string().c_str());
}

void capture_data(SNDFILE * output_stream, uint32_t decimation) {
	stream_cmd_t stream_cmd(stream_cmd_t::STREAM_MODE_START_CONTINUOUS);
	rx_metadata_t metadata;
	
	size_t received_samples;
	sf_count_t samples_written;
	uint32_t i;
	uint32_t j;
	uint32_t start_point;

	double * c1_filtered;
	double * c2_filtered;
	
	float * audio;
	
	double timeout;
	
	std::vector<std::vector<std::complex<float> > > buffers(
		dev->get_rx_num_channels(), std::vector<std::complex<float> >
		(BUFFER_SIZE));
	
	//create a vector of pointers to point to each of the channel buffers
	std::vector<std::complex<float> *> buffer_ptrs;
	for (i = 0; i < buffers.size(); i++) 
	{
		buffer_ptrs.push_back(&buffers[i].front());
	}
	
	uhd::stream_args_t stream_args("fc32", "sc16");
	
	for (i = 0; i < dev->get_rx_num_channels(); i++) 
	{
		stream_args.channels.push_back(i); //linear mapping
	}
	
	uhd::rx_streamer::sptr rx_stream = dev->get_rx_stream(stream_args);
	
	
	stream_cmd.num_samps = 0;
	stream_cmd.stream_now = false;
	stream_cmd.time_spec = uhd::time_spec_t(STARTUP_TIME);
	
	audio = (float *) malloc(sizeof(float) * BUFFER_SIZE * 2);
	c1_filtered = (double *) malloc(sizeof(double) * BUFFER_SIZE);
	c2_filtered = (double *) malloc(sizeof(double) * BUFFER_SIZE);
	
	
	if (audio == 0x00) 
	{
		syslog(LOG_ERR, "Couldn't allocate buffer for audio! Quitting.");
		exit(EXIT_FAILURE);
	}
	
	if ((c1_filtered == 0x00) || (c2_filtered == 0x00)) 
	{
		syslog(LOG_ERR, "Couldn't allocate buffer for filtered samples! Quitting.");
		exit(EXIT_FAILURE);
	}

	
	
	
	dev->issue_stream_cmd(stream_cmd);
	
	timeout = STARTUP_TIME + USRP_TIMEOUT;	// Allow for the startup time.
	
	start_point = 0;
	
	while (1) 
	{
		received_samples = rx_stream->recv(buffer_ptrs, BUFFER_SIZE, metadata, timeout);

		
		if (metadata.error_code != rx_metadata_t::ERROR_CODE_NONE)
		{
			syslog(LOG_ERR, "Received metadata error code 0x%X. Quitting.", metadata.error_code);
			exit(EXIT_FAILURE);
		}
		
		
		if (received_samples != BUFFER_SIZE)
		{
			syslog(LOG_WARNING, "Received less samples than requested. Wanted %d, got %d.", BUFFER_SIZE, (int) received_samples);
		}
		
		timeout = USRP_TIMEOUT;	// Reset to the normal value.
		
		
		
		// Process the data
			
		// Filter channels
		
		for (i = 0; i < received_samples; i++)
		{
			c1_filtered[i] = filter(real(buffers[0][i]) * SCALING_GAIN, 0);
			c2_filtered[i] = filter(real(buffers[1][i]) * SCALING_GAIN, 1);
		}
	
		// Decimate
		
		j = 0;
		
		for (i = start_point; i < received_samples; i += decimation)
		{
			audio[j] = (float) c1_filtered[i];
			
			j++;
			
			audio[j] = (float) c2_filtered[i];
			
			j++;
		}
		
		start_point = ((i + decimation) - received_samples);
		
		
		
		
		// Write out the audio
		
		
		samples_written = sf_write_float(output_stream, audio, j);
		
		if (samples_written != j)
		{
			syslog(LOG_ERR, "Short write to the audio file!");
			
			sf_close(output_stream);
			
			exit(EXIT_FAILURE);
		}
	}
}

void my_uhd_handler(uhd::msg::type_t type, const std::string &msg)
{	
	if (type == msg::type_t::status)
	{
		syslog(LOG_DEBUG, "%s", msg.c_str());
	}
	else if (type == msg::type_t::warning)
	{
		syslog(LOG_WARNING, "%s", msg.c_str()); 
	}
	else if(type == msg::type_t::error)
	{
		syslog(LOG_ERR, "%s", msg.c_str());
	}
	else 
	{
		syslog(LOG_DEBUG, "%s", msg.c_str());
	}
}

double filter(double input, uint32_t channel)
{
	uint32_t i;
	double output;
	
	static double filter_state_c1[BL];
	static double filter_state_c2[BL];
	
	if (channel == 0)
	{
		for (i = BL-1; i > 0; i--)
		{
			filter_state_c1[i] = filter_state_c1[i-1];
		}
		
		filter_state_c1[0] = input;
		
		output = 0.0;
		
		for (i = 0; i < BL; i++) {
			output += B[i] * filter_state_c1[i];
		}
	}
	else
	{
		for (i = BL-1; i > 0; i--)
		{
			filter_state_c2[i] = filter_state_c2[i-1];
		}
		
		filter_state_c2[0] = input;
		
		output = 0.0;
		
		for (i = 0; i < BL; i++) {
			output += B[i] * filter_state_c2[i];
		}
	}
	
	return output;
}
