////////////////////
// telescope_dump //
////////////////////

/******************************************************************************
 * Copyright  2011, 2012 Mike Roddewig (mroddewi@mines.edu).
 * All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License v3 as published 
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
******************************************************************************/

#include <syslog.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include <stdio.h>
#include <argp.h>
#include <unistd.h>
#include <complex>

#include "uhd/types/device_addr.hpp"
#include "uhd/types/ref_vector.hpp"
#include "uhd/usrp/multi_usrp.hpp"
#include "uhd/device.hpp"
#include <uhd/utils/thread_priority.hpp>
#include <uhd/utils/safe_main.hpp>
#include <uhd/utils/msg.hpp>

#include <boost/thread.hpp>

#include "telescope.h"
#include "version.h"

using namespace uhd;
using namespace uhd::usrp;

/////////////
// Defines //
/////////////

#define USRP_TIMEOUT		0.1	// sec
#define USRP_SAMPLE_RATE	5E6	// SPS
#define USRP_RX_FREQ		20.1E6	// Hz
#define USRP_GAIN		3	// dB
#define STARTUP_TIME		1.5	// sec
#define BUFFER_SIZE		1024	// samples
#define INTEGRATE_DATASETS	100	// sets of received samples

////////////////
// argp stuff //
////////////////

const char * program_name = "telescope_dump";

/* This structure is used by main to communicate with parse_opt. */
struct arguments {
	char * address;
	uint8_t foreground;
	uint8_t verbose;
	double sample_rate;
	double rx_freq;
	double gain;
};


const char *argp_program_version = TELESCOPE_DUMP_VERSION;

const char *argp_program_bug_address = "Mike Roddewig <mroddewi@mines.edu>";

/*
   OPTIONS.  Field 1 in ARGP.
   Order of fields: {NAME, KEY, ARG, FLAGS, DOC}.
*/
static struct argp_option options[] = {
	{"foreground", 'f', 0, 0,
		"Keep running in the foreground instead of forking off into the background."},
	{"samplerate", 's', "SAMPLERATE", OPTION_ARG_OPTIONAL, 
		"Receiver sample rate after decimation, in SPS. Default is 5 MSPS."},
	{"rxfreq", 'r', "RXFREQ", OPTION_ARG_OPTIONAL,
		"Receiver frequency to tune to, in Hz. Default is 20.1 MHz."},
	{"gain", 'g', "GAIN", OPTION_ARG_OPTIONAL,
		"Receiver gain in dB. Default is 3 dB."},
	{"verbose", 'v', 0, 0,
		"Turn on extra messages."},
	{0}
};


/*
   PARSER. Field 2 in ARGP.
   Order of parameters: KEY, ARG, STATE.
*/

static error_t parse_opt (int key, char *arg, struct argp_state *state) {
	struct arguments *arguments = (struct arguments *) state->input;
	int ret;

	switch (key) 
	{
		case ARGP_KEY_ARG:
                        if (state->arg_num >= 1) {
                                argp_usage(state);
                        }
                        arguments->address = arg;
                        break;
		case 'f':
			arguments->foreground = 1;
			break;
		case 'v':
			arguments->verbose = 1;
			break;
		case 's':
			ret = sscanf(arg, "%lf", &arguments->sample_rate);
			if (ret != 1)
			{
				return ARGP_ERR_UNKNOWN;
			}
			break;
		case 'g':
			ret = sscanf(arg, "%lf", &arguments->gain);
			if (ret != 1)
			{
				return ARGP_ERR_UNKNOWN;
			}
			break;
		case 'r':
			ret = sscanf(arg, "%lf", &arguments->rx_freq);
			if (ret != 1)
			{
				return ARGP_ERR_UNKNOWN;
			}
			break;
		case ARGP_KEY_END:
			break;
		default:
			return ARGP_ERR_UNKNOWN;
	}
	
	return 0;
}

/*
   ARGS_DOC. Field 3 in ARGP.
   A description of the non-option command-line arguments
     that we accept.
*/
static char args_doc[] = "[ADDRESS]";

/*
  DOC.  Field 4 in ARGP.
  Program documentation.
*/
static char doc[] = "Capture data from the USRP at ADDRESS.";

/*
   The ARGP structure itself.
*/
static struct argp argp = {options, parse_opt, args_doc, doc};

///////////////////////
// Private Functions //
///////////////////////

void initialize_usrp(char *, double, double, double);
void capture_data(FILE *);
void my_uhd_handler(uhd::msg::type_t, const std::string &);

//////////////////////
// Global Variables //
//////////////////////

multi_usrp::sptr dev;
double sample_time;

//////////
// Main //
//////////

int UHD_SAFE_MAIN(int argc, char * argv[]) 
{	
	struct arguments arguments;
	
	FILE * output_stream;
	
	pid_t pid;
	pid_t sid;
	
	uhd::msg::register_handler(&my_uhd_handler); // Redirect UHD messages to syslog
	
	set_thread_priority_safe();

	output_stream = stdout;
	
	// Handle any command-line arguments.
	
	arguments.address = NULL;
	arguments.foreground = 0;
	arguments.rx_freq = 0;
	arguments.sample_rate = 0;
	arguments.gain = 0;
	arguments.verbose = 0;
	
	argp_parse(&argp, argc, argv, 0, 0, &arguments);
	
	if (arguments.rx_freq == 0)
	{
		arguments.rx_freq = USRP_RX_FREQ;
	}
	
	if (arguments.sample_rate == 0)
	{
		arguments.sample_rate = USRP_SAMPLE_RATE;
	}
	
	if (arguments.gain == 0)
	{
		arguments.gain = USRP_GAIN;
	}
	
	if (arguments.foreground != 1) {
		// Daemonize ourselves.
		
		pid = fork();
		
		if (pid < 0) {
			fprintf(stderr, "Couldn't daemonize!\n");
			exit(EXIT_FAILURE);
		}
		
		if (pid > 0) {
			exit(EXIT_SUCCESS);
		}
		
		sid = setsid();
		if (sid < 0) {
			syslog(LOG_ERR, "couldn't get an SID: %s. Quitting.\n", strerror(errno));
			exit(EXIT_FAILURE);
		}
		
		openlog(program_name, 0, LOG_DAEMON);
	}
	else {
		openlog(program_name, LOG_PERROR, LOG_USER);
	}

	syslog(LOG_INFO, "starting up.");
	
	if (arguments.verbose == 0)
	{
		setlogmask(~(LOG_MASK(LOG_DEBUG)));
	}
	
	if (arguments.address == NULL) {
		syslog(LOG_ERR, "you must specify the address of the USRP. Quitting.");
		exit(EXIT_FAILURE);
	}
	
	sample_time = 1/((double) arguments.sample_rate);
	
	initialize_usrp(arguments.address, arguments.sample_rate, 
		arguments.rx_freq, arguments.gain);
	capture_data(output_stream);
	
	exit(EXIT_SUCCESS);
}

void initialize_usrp(char * address, double sample_rate, double rx_freq, double gain)
{
	device_addr_t dev_addr;
	tune_request_t tune_request;
	tune_result_t tune_result;
	
	dev_addr["addr0"] = address;
	
	dev = multi_usrp::make(dev_addr);
	
	dev->set_rx_subdev_spec(subdev_spec_t("A:A A:B"));
	dev->set_rx_rate(sample_rate, multi_usrp::ALL_CHANS);
	
	syslog(LOG_DEBUG, "Actual receive rate: %f S/s", dev->get_rx_rate(0));
	
	dev->set_rx_gain(gain, multi_usrp::ALL_GAINS);
	
	syslog(LOG_DEBUG, "Actual gain: %f dB", dev->get_rx_gain(0));
	
	dev->set_rx_dc_offset(false, multi_usrp::ALL_CHANS);
	
	dev->set_time_now(uhd::time_spec_t(0.0));
        boost::this_thread::sleep(boost::posix_time::milliseconds(50));	// wait for time sync
	
	tune_request = tune_request_t(rx_freq);
	tune_result = dev->set_rx_freq(tune_request, 0);
	
	syslog(LOG_DEBUG, "Actual DSP0 frequency: %f Hz", tune_result.actual_dsp_freq);
	
	tune_result = dev->set_rx_freq(tune_request, 1);
	
	syslog(LOG_DEBUG, "Actual DSP1 frequency: %f Hz", tune_result.actual_dsp_freq);
}

void capture_data(FILE * output_stream) {
	stream_cmd_t stream_cmd(stream_cmd_t::STREAM_MODE_START_CONTINUOUS);
	rx_metadata_t metadata;
	
	size_t received_samples;
	uint32_t i;
	int ret;
	
	double timeout;
	
	double sample;
	
	std::vector<std::vector<std::complex<float> > > buffers(
		dev->get_rx_num_channels(), std::vector<std::complex<float> >
		(BUFFER_SIZE));
	
	//create a vector of pointers to point to each of the channel buffers
	std::vector<std::complex<float> *> buffer_ptrs;
	for (i = 0; i < buffers.size(); i++) 
	{
		buffer_ptrs.push_back(&buffers[i].front());
	}
	
	uhd::stream_args_t stream_args("fc32");
	
	for (i = 0; i < dev->get_rx_num_channels(); i++) 
	{
		stream_args.channels.push_back(i); //linear mapping
	}
	
	uhd::rx_streamer::sptr rx_stream = dev->get_rx_stream(stream_args);
	
	
	
	stream_cmd.num_samps = 0;
	stream_cmd.stream_now = false;
	stream_cmd.time_spec = uhd::time_spec_t(STARTUP_TIME);
	
	dev->issue_stream_cmd(stream_cmd);
	
	timeout = STARTUP_TIME + USRP_TIMEOUT;	// Allow for the startup time.
	
	while (1) 
	{
		
		received_samples = rx_stream->recv(buffer_ptrs, BUFFER_SIZE, metadata, timeout);
		
		if (metadata.error_code != rx_metadata_t::ERROR_CODE_NONE)
		{
			syslog(LOG_ERR, "Received metadata error code 0x%X. Quitting.", metadata.error_code);
			exit(EXIT_FAILURE);
		}
		
		
		
		
		// Write the samples
		
		for (i = 0; i < received_samples; i++)
		{
			sample = real(buffers[0][i]);

			ret = fwrite(&sample, sizeof(double), 1, output_stream);
			if (ret != 1) 
			{
				syslog(LOG_ERR, "short write. Quitting. Error: %s.", strerror(errno));
				exit(EXIT_FAILURE);
			}
			
			sample = real(buffers[1][i]);

			ret = fwrite(&sample, sizeof(double), 1, output_stream);
			if (ret != 1) 
			{
				syslog(LOG_ERR, "short write. Quitting. Error: %s.", strerror(errno));
				exit(EXIT_FAILURE);
			}
		}
		
		fflush(output_stream);
		
		timeout = USRP_TIMEOUT;	// Reset to the normal value.
	}
}

void my_uhd_handler(uhd::msg::type_t type, const std::string &msg)
{	
	if (type == msg::type_t::status)
	{
		syslog(LOG_DEBUG, "%s", msg.c_str());
	}
	else if (type == msg::type_t::warning)
	{
		syslog(LOG_WARNING, "%s", msg.c_str()); 
	}
	else if(type == msg::type_t::error)
	{
		syslog(LOG_ERR, "%s", msg.c_str());
	}
	else 
	{
		syslog(LOG_DEBUG, "%s", msg.c_str());
	}
}
