//////////////////////
// The File Process //
//////////////////////

/******************************************************************************
 * The file process reads data from the compressor process and writes it out to
 * the appropriately named file. It also manages the creation of the year/
 * month/day/hour directory structure.
******************************************************************************/

/******************************************************************************
 * Copyright  2011, Mike Roddewig (mroddewi@mines.edu).
 * All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License v3 as published 
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
******************************************************************************/

#include <stdio.h>
#include <time.h>
#include <errno.h>
#include <string.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <stdlib.h>
#include <dirent.h>
#include <unistd.h>
#include <stdint.h>
#include <argp.h>
#include <assert.h>

#include "telescope.h"
#include "version.h"

/////////////
// Defines //
/////////////

#define BUFFER_SIZE 	1024
#define RECORDS_MODE	(S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH)

////////////////
// argp stuff //
////////////////

const char * program_name = "telescope_filed";

/* This structure is used by main to communicate with parse_opt. */
struct arguments {
	char * records_directory;
	char * input;
	uint8_t foreground;
	uint8_t flush;
};


const char *argp_program_version = TELESCOPE_FILED_VERSION;

const char *argp_program_bug_address = "Mike Roddewig <mroddewi@mines.edu>";

/*
   OPTIONS.  Field 1 in ARGP.
   Order of fields: {NAME, KEY, ARG, FLAGS, DOC}.
*/
static struct argp_option options[] = {
	{"foreground", 'f', 0, 0,
		"Keep running in the foreground instead of forking off into the background."},
	{"flush", 'w', 0, 0,
		"Flush the output stream after every sample."},
	{"input", 'i', "INPUT", 0,
		"Input file, usually a FIFO."},
	{0}
};


/*
   PARSER. Field 2 in ARGP.
   Order of parameters: KEY, ARG, STATE.
*/

static error_t parse_opt (int key, char *arg, struct argp_state *state) {
	struct arguments *arguments = (struct arguments *) state->input;

	switch (key) {
		case ARGP_KEY_ARG:
			if (state->arg_num >= 1) {
				argp_usage(state);
			}
			arguments->records_directory = arg;
			break;
		case 'i':
			arguments->input = arg;
			break;
		case 'f':
			arguments->foreground = 1;
			break;
		case 'w':
			arguments->flush = 1;
			break;
		case ARGP_KEY_END:
			break;
		default:
		return ARGP_ERR_UNKNOWN;
	}
	
	return 0;
}

/*
   ARGS_DOC. Field 3 in ARGP.
   A description of the non-option command-line arguments
     that we accept.
*/
static char args_doc[] = "RECORD_DIRECTORY";

/*
  DOC.  Field 4 in ARGP.
  Program documentation.
*/
static char doc[] = "Save the recordings to RECORD_DIRECTORY.";

/*
   The ARGP structure itself.
*/
static struct argp argp = {options, parse_opt, args_doc, doc};

///////////////////////
// Private Functions //
///////////////////////

FILE * get_current_file(struct tm *, double, char *);
void update_directory(struct tm *, int);

//////////////////////
// Global Variables //
//////////////////////

uint8_t flush;

int main(int argc, char * argv[]) {
	struct arguments arguments;
	
	FILE * data_file;
	int records_dirfd;
	
	FILE * input_stream;
	
	int ret;
	
	int current_day;
	
	pid_t pid;
	pid_t sid;
	
	char file_buffer[BUFFER_SIZE];
	
	telescope_message_header_t header;
	
	// Handle any command-line arguments.
	
	arguments.records_directory = NULL;
	arguments.input = NULL;
	arguments.foreground = 0;
	arguments.flush = 0;
	
	argp_parse(&argp, argc, argv, 0, 0, &arguments);
	
	if (arguments.foreground != 1) {
		// Daemonize ourselves.
		
		pid = fork();
		
		if (pid < 0) {
			fprintf(stderr, "Couldn't daemonize!\n");
			exit(EXIT_FAILURE);
		}
		
		if (pid > 0) {
			exit(EXIT_SUCCESS);
		}
		
		sid = setsid();
		if (sid < 0) {
			syslog(LOG_ERR, "couldn't get an SID: %s. Quitting.\n", strerror(errno));
			exit(EXIT_FAILURE);
		}
		
		openlog(program_name, 0, LOG_DAEMON);
	}
	else {
		openlog(program_name, LOG_PERROR, LOG_USER);
	}
	
	flush = arguments.flush;

	syslog(LOG_INFO, "starting up.\n");
	
	setlogmask(~(LOG_MASK(LOG_DEBUG)));
	
	if (arguments.records_directory == NULL) {
		syslog(LOG_ERR, "you must specify a recording directory. Quitting.\n");
		exit(EXIT_FAILURE);
	}
	
	if (arguments.input == NULL)
	{
		syslog(LOG_ERR, "you must specify an input file. Quitting.");
		exit(EXIT_FAILURE);
	}
	
	input_stream = fopen(arguments.input, "r");
	
	if (input_stream == NULL)
	{
		syslog(LOG_ERR, "Couldn't open the input file %s: %s", arguments.input, strerror(errno));
		exit(EXIT_FAILURE);
	}
	
	// Do some more daemon specific things.
	
	ret = chdir(arguments.records_directory);
	
	if (ret != 0) {
		syslog(LOG_ERR, "couldn't change the working directory to %s: %s.\n", 
			arguments.records_directory, strerror(errno));
		exit(EXIT_FAILURE);
	}
	
	records_dirfd = dirfd(opendir(arguments.records_directory));
        if (records_dirfd == -1) {
                syslog(LOG_ERR, "failure opening the records directory: %s. Quitting.\n", strerror(errno));
                exit(EXIT_FAILURE);
        }
	
	current_day = -1;
	
	while (1) {
		
		ret = fread(&header, sizeof(telescope_message_header_t), 1, input_stream);
		if (ret != 1) {
			syslog(LOG_ERR, "got a short read on the message header. Received %d bytes. Quitting.", ret);
			exit(EXIT_FAILURE);
		}
		
		assert(header.num_samples > 0);

		ret = fread(&file_buffer, sizeof(double), header.num_samples, input_stream);
		if (ret != header.num_samples) {
			syslog(LOG_ERR, "got a short read on the message data. Received %d samples, expected %d. Quitting.", 
				ret, header.num_samples);
			exit(EXIT_FAILURE);
		}
		
		if ((current_day == -1) || (current_day != header.record_time.tm_mday)) 
		{
			if ((current_day != -1) && (data_file != 0x00)) 
			{
				fclose(data_file);
			}
			
			current_day = header.record_time.tm_mday;
			
			update_directory(&header.record_time, records_dirfd);
			data_file = get_current_file(&header.record_time, header.fractional_secs, arguments.records_directory);
		}
		
		ret = fwrite(&file_buffer, sizeof(double), header.num_samples, data_file);
		if (ret != header.num_samples) {
			syslog(LOG_ERR, "file: error writing the data to the output file: %s. Quitting.\n", strerror(errno));
			exit(EXIT_FAILURE);
		}
		
		if (flush != 0)
		{
			fflush(data_file);
		}
	}
}



// Creates a year/month directory structure for time utc_time beneath the given
// records_directory. Kills the program if it encounters any filesystem errors.
void update_directory(struct tm * utc_time, int records_directory) {
	uint8_t ret;
	char path_name[64];
	int year;
	int month;
	
	year = utc_time->tm_year + 1900;
	month = utc_time->tm_mon + 1;
	
	// Create the year directory.
	
	sprintf(path_name, "%d", year);
	ret = mkdirat(records_directory, path_name, RECORDS_MODE);
	
	if ((ret != 0) && (errno != EEXIST)) {
		syslog(LOG_ERR, "file: error creating the year directory for the recordings: %s.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	
	// Create the month directory.
	
	sprintf(path_name, "%d/%02d", year, month);
	ret = mkdirat(records_directory, path_name, RECORDS_MODE);
	
	if ((ret != 0) && (errno != EEXIST)) {
		syslog(LOG_ERR, "file: error creating the month directory for the recordings: %s.\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
}

// Creates the minute.bin file under the records directory in the appropriate year/month/ directory.
// Will kill the program if it encounters any filesystem errors. 
FILE * get_current_file(struct tm * utc_time, double fractional_secs, char * records_directory) {
	char * file_name;
	int year;
	int month;
	FILE * fd;
	double seconds;
	
	year = utc_time->tm_year + 1900;
	month = utc_time->tm_mon + 1;
	
	seconds = utc_time->tm_sec + fractional_secs;
	
	file_name = (char *) malloc(strlen(records_directory) + 64);
	
	sprintf(file_name, "%s/%d/%02d/%d-%02d-%02d-%02d:%02d:%.4f.bin", records_directory, year, month, 
					year, month, utc_time->tm_mday, utc_time->tm_hour, utc_time->tm_min, seconds);
	
	fd = fopen(file_name, "w+");
	
	if (fd != 0) {
		free(file_name);
		return fd;
	}
	else {
		syslog(LOG_ERR, "Error opening the records file %s: %s. Quitting.", file_name, strerror(errno));
		free(file_name);
		exit(EXIT_FAILURE);
	}
}